#include "dt_blob.h"
#include "mat.h"

/*
  classifiers data structure:
     classifiers(1).data_length (double)
     classifiers(1).window_length (double)
     classifiers(1).fft_params.{fs (double), band_boundaries (17x1 double), 
                                       window_size (double), overlapping_windows (double)}
     classifiers(*).threshold (double)
     classifiers(*).features_used (Fx1 char)
     classifiers(*).wcs (1xnum_wc struct)
        wcs(*).node_confidences (Nx1 double)
	wcs(*).var (Nx1 int32)
	wcs(*).cut (Nx1 double)
	wcs(*).left_child (Nx1 int32)
	wcs(*).right_child (Nx1 int32)
	wcs(*).cat_split (Mx1 cell - double arrays)
  
  lay out in order listed in contiguous block, placing array lengths just before data when necessary
*/

void*
dt2blob(mxArray *classifiers_mx, int* numbytes) {


  int blobsize = sizeof(double)*5+sizeof(int);

  int num_classifiers = mxGetNumberOfElements(classifiers_mx);

  mxArray* data_length_mx = mxGetField(classifiers_mx, 0, "data_length");
  double data_length = *(*double)mxGetData(data_length_mx);
  
  mxArray* window_length_mx = mxGetField(classifiers_mx, 0, "window_length");
  double window_length = *(*double)mxGetData(window_length_mx);
  
  mxArray* fft_params_mx = mxGetField(classifiers_mx, 0, "fft_params");
  
  mxArray* fs_mx = mxGetField(fft_params_mx, 0, "fs");
  double fs = *(*double)mxGetData(fs_mx);
  mxArray* bb_mx = mxGetField(fft_params_mx, 0, "band_boundaries");
  int num_band_boundaries = mxGetNumberOfElements(bb_mx); 
  double* band_boundaries = (*double)mxGetData(bb_mx);
  mxArray* ws_mx = mxGetField(fft_params_mx, 0, "window_size");
  double window_size = *(*double)mxGetData(ws_mx);
  mxArray* ow_mx = mxGetField(fft_params_mx, 0, "overlapping_windows");
  double overlapping_windows = *(*double)mxGetData(ow_mx);

  blobsize += sizeof(double)*num_bands + sizeof(int);

  mxArray* wcs_mx;

  /* count blob size here */
  int c = 0;
  for (c = 0; c < num_classifiers; c++) {

    blobsize += sizeof(double); /*threshold*/

    int num_features = mxGetNumberOfElements(mxGetField(classifiers_mx, c, "features_used"));
    blobsize += sizeof(int) + num_features*sizeof(char);

    wcs_mx = mxGetField(classifiers_mx, c, "wcs");
    int num_wcs = mxGetNumberOfElements(wcs_mx);
    blobsize += sizeof(int);
    int w = 0;
    for (w = 0; w < num_wcs; w++) {
      int num_nodes = mxGetNumberOfElements(mxGetField(wcs_mx, w, "node_confidences"));
      blobsize += num_nodes*(2*sizeof(double)+3*sizeof(int)) + sizeof(int);

      catsplit_mx = mxGetField(wcs_mx, w, "catsplit");
      int nsplits = mxGetNumberOfElements(catsplit_mx);
      blobsize += sizeof(int);

      int n = 0;
      for (n = 0; n < nsplits; n++) {
	mxArray* catsplit_cell_mx = mxGetCell(catsplit_mx, n);
	int ncatsplit = mxGetNumberOfElements(catsplit_cell_mx);
	blobsize += sizeof(int) + sizeof(double)*ncatsplit;
      }
    }
  }


  /* create contiguous blob of size blobsize */
  char* blob = (char*)malloc(blobsize);
  int currpos = 0;
  memcpy((void*)&blob[currpos], (void*)&num_classifiers, sizeof(int));
  currpos += sizeof(int);
  memcpy((void*)&blob[currpos], (void*)&data_length, sizeof(double));
  currpos += sizeof(double); 
  memcpy((void*)&blob[currpos], (void*)&window_length, sizeof(double));
  currpos += sizeof(double); 
  memcpy((void*)&blob[currpos], (void*)&fs, sizeof(double));
  currpos += sizeof(double); 
  memcpy((void*)&blob[currpos], (void*)&num_band_boundaries, sizeof(int));
  currpos += sizeof(int);
  memcpy((void*)&blob[currpos], (void*)band_boundaries, sizeof(double)*num_band_boundaries);
  currpos += sizeof(double)*num_band_boundaries; 
  memcpy((void*)&blob[currpos], (void*)&window_size, sizeof(double));
  currpos += sizeof(double); 
  memcpy((void*)&blob[currpos], (void*)&overlapping_windows, sizeof(double));
  currpos += sizeof(double); 
  
  for (c = 0; c < num_classifiers; c++) {
    memcpy((void*)&blob[currpos], mxGetData(mxGetField(classifiers_mx, c, "threshold")), sizeof(double));
    currpos += sizeof(double);
    mxArray* features_used_mx = mxGetField(classifiers_mx, c, "features_used");
    int num_features = mxGetNumberOfElements(features_used_mx);
    memcpy((void*)&blob[currpos], (void*)&num_features, sizeof(int));
    currpos += sizeof(int);
    memcpy((void*)&blob[currpos], mxGetData(features_used_mx), sizeof(char)*num_features);
    currpos += sizeof(char)*num_features;
    
    wcs_mx = mxGetField(classifiers_mx, c, "wcs");
    int num_wcs = mxGetNumberOfElements(wcs_mx);
    memcpy((void*)&blob[currpos], (void*)&num_wcs, sizeof(int));
    currpos += sizeof(int);   

    for (w = 0; w < num_wcs; w++) {
      int num_nodes = mxGetNumberOfElements(mxGetField(wcs_mx, w, "node_confidences"));
      memcpy((void*)&blob[currpos], (void*)&num_nodes, sizeof(int));
      currpos += sizeof(int);   
      memcpy((void*)&blob[currpos], mxGetData(mxGetField(wcs_mx, w, "node_confidences")), sizeof(double)*num_nodes);
      currpos += sizeof(double)*num_nodes;
      memcpy((void*)&blob[currpos], mxGetData(mxGetField(wcs_mx, w, "var")), sizeof(int)*num_nodes);
      currpos += sizeof(int)*num_nodes;
      memcpy((void*)&blob[currpos], mxGetData(mxGetField(wcs_mx, w, "cut")), sizeof(double)*num_nodes);
      currpos += sizeof(double)*num_nodes;
      memcpy((void*)&blob[currpos], mxGetData(mxGetField(wcs_mx, w, "left_child")), sizeof(int)*num_nodes);
      currpos += sizeof(int)*num_nodes;
      memcpy((void*)&blob[currpos], mxGetData(mxGetField(wcs_mx, w, "right_child")), sizeof(int)*num_nodes);
      currpos += sizeof(int)*num_nodes;

      catsplit_mx = mxGetField(wcs_mx, w, "catsplit");
      int nsplits = mxGetNumberOfElements(catsplit_mx);
      memcpy((void*)&blob[currpos], (void*)&nsplits, sizeof(int));
      currpos += sizeof(int);         
      int n = 0;
      for (n = 0; n < nsplits; n++) {
	mxArray* catsplit_cell_mx = mxGetCell(catsplit_mx, n);
	int ncatsplit = mxGetNumberOfElements(catsplit_cell_mx);
	memcpy((void*)&blob[currpos], (void*)&ncatsplit, sizeof(int));
	currpos += sizeof(int);         
	memcpy((void*)&blob[currpos], mxGetData(catsplit_cell_mx), sizeof(double)*ncatsplit);
	currpos += sizeof(double)*ncatsplit;
      }
    }  
  }
  printf("%d   %d\n", blobsize, currpos);
  
  *numbytes = currpos;
  
  return (void*)blob;
  
}  

/*
  classifiers data structure:
     classifiers(1).data_length (double) 
     classifiers(1).window_length (double) 
     classifiers(1).fft_params.{fs (double), band_boundaries (17x1 double),  
                                       window_size (double), overlapping_windows (double)} 
     classifiers(*).threshold (double) 
     classifiers(*).features_used (Fx1 char) 
     classifiers(*).wcs (1xnum_wc struct) 
        wcs(*).node_confidences (Nx1 double) 
	wcs(*).var (Nx1 int32) 
	wcs(*).cut (Nx1 double) 
	wcs(*).left_child (Nx1 int32) 
	wcs(*).right_child (Nx1 int32) 
	wcs(*).cat_split (Mx1 cell - double arrays)
	
  array lengths are stored immediately before array data 
*/
mxArray* 
blob2dt(char* blob, int numbytes) {
  
  int currpos = 0;
  num_classifiers = *(int*)(&blob[currpos]);  
  currpos += sizeof(int);

  /* general info */
  mxArray* classifiers_mx = mxCreateStructArray(2, {num_classifiers, 1}, 6, 
      {"data_length, window_length, fft_params, threshold, features_used, wcs"});
  mxArray* tmp = mxCreateNumericArray(2, {1, 1}, mxDOUBLE_CLASS, mxREAL);
  mxSetData(tmp, (void*)&blob[currpos]);
  mxSetField(classifiers_mx, 0, "data_length", tmp);
  currpos += sizeof(double);
  tmp = mxCreateNumericArray(2, {1, 1}, mxDOUBLE_CLASS, mxREAL);
  mxSetData(tmp, (void*)&blob[currpos]);
  mxSetField(classifiers_mx, 0, "window_length", tmp);
  currpos += sizeof(double);
  
  /* fft_params */
  mxArray* fft_params_mx = mxCreateStructArray(2, {1, 1}, 4, 
      {"fs", "band_boundaries", "window_size", "overlapping_windows"});
  tmp = mxCreateNumericArray(2, {1, 1}, mxDOUBLE_CLASS, mxREAL);
  mxSetData(tmp, (void*)&blob[currpos]);
  mxSetField(fft_params_mx, 0, "fs", tmp);
  currpos += sizeof(double);
  int num_band_boundaries = *(int*)(&blob[currpos]);
  currpos += sizeof(int);
  tmp = mxCreateNumericArray(2, {num_band_boundaries, 1}, mxDOUBLE_CLASS, mxREAL);
  mxSetData(tmp, (void*)&blob[currpos]);
  mxSetField(fft_params_mx, 0, "band_boundaries", tmp);
  currpos += sizeof(double)*num_band_boundaries;
  tmp = mxCreateNumericArray(2, {1, 1}, mxDOUBLE_CLASS, mxREAL);
  mxSetData(tmp, (void*)&blob[currpos]);
  mxSetField(fft_params_mx, 0, "window_size", tmp);
  currpos += sizeof(double);
  tmp = mxCreateNumericArray(2, {1, 1}, mxDOUBLE_CLASS, mxREAL);
  mxSetData(tmp, (void*)&blob[currpos]);
  mxSetField(fft_params_mx, 0, "overlapping_windows", tmp);
  currpos += sizeof(double);
  mxSetField(classifiers_mx, 0, "fft_params", fft_params_mx);

  mxArray* wcs_mx;

  /* classifier info */  
  for (c = 0; c < num_classifiers; c++) {

    tmp = mxCreateNumericArray(2, {1, 1}, mxDOUBLE_CLASS, mxREAL);
    mxSetData(tmp, (void*)&blob[currpos]);
    mxSetField(classifiers_mx, c, "threshold", tmp);
    currpos += sizeof(double);
    int num_features_used = *(int*)(&blob[currpos]);
    currpos += sizeof(int);
    tmp = mxCreateNumericArray(2, {num_features_used, 1}, mxUINT8_CLASS, mxREAL);
    mxSetData(tmp, (void*)&blob[currpos]);
    mxSetField(classifiers_mx, c, "features_used", tmp);
    currpos += sizeof(char)*num_features_used;

    /* weak classifiers info */
    int num_wcs = *(int*)(&blob[currpos]);
    currpos += sizeof(int);
    
    wcs_mx = mxCreateStructArray(2, {num_wcs, 1}, 6, 
	{"node_confidences", "var", "cut", "left_child", "right_child", "catsplit"});

    for (w = 0; w < num_wcs; w++) {
      int num_nodes = *(int*)(&blob[currpos]);
      currpos += sizeof(int);

      tmp = mxCreateNumericArray(2, {num_nodes, 1}, mxDOUBLE_CLASS, mxREAL);
      mxSetData(tmp, (void*)&blob[currpos]);
      mxSetField(wcs_mx, w, "node_confidences", tmp);
      currpos += sizeof(double)*num_nodes;
      tmp = mxCreateNumericArray(2, {num_nodes, 1}, mxINT32_CLASS, mxREAL);
      mxSetData(tmp, (void*)&blob[currpos]);
      mxSetField(wcs_mx, w, "var", tmp);
      currpos += sizeof(int)*num_nodes;
      tmp = mxCreateNumericArray(2, {num_nodes, 1}, mxDOUBLE_CLASS, mxREAL);
      mxSetData(tmp, (void*)&blob[currpos]);
      mxSetField(wcs_mx, w, "cut", tmp);
      currpos += sizeof(double)*num_nodes;
      tmp = mxCreateNumericArray(2, {num_nodes, 1}, mxINT32_CLASS, mxREAL);
      mxSetData(tmp, (void*)&blob[currpos]);
      mxSetField(wcs_mx, w, "left_child", tmp);
      currpos += sizeof(int)*num_nodes;
      tmp = mxCreateNumericArray(2, {num_nodes, 1}, mxINT32_CLASS, mxREAL);
      mxSetData(tmp, (void*)&blob[currpos]);
      mxSetField(wcs_mx, w, "right_child", tmp);
      currpos += sizeof(int)*num_nodes;

      int nsplits = *(int*)(&blob[currpos]);
      currpos += sizeof(int);
      mxArray* catsplit_mx = mxCreateCellArray(2, {nsplits, 1});
      
      int n = 0;
      for (n = 0; n < nsplits; n++) {
	int ncatsplit = *(int*)(&blob[currpos]);
	currpos += sizeof(int);
	tmp = mxCreateNumericArray(2, {ncatsplit, 1}, mxDOUBLE_CLASS, mxREAL);
	mxSetData(tmp, (void*)&blob[currpos]);
	mxSetCell(catsplit_mx, n, tmp);
	currpos += sizeof(double)*ncatsplit;
      }
      mxSetField(wcs_mx, w, "catsplit", catsplit_mx);
      
    }  
    
    mxSetField(classifiers_mx, c, "wcs", wcs_mx);

  }

  printf("%d   %d\n", numbytes, currpos);

  return classifiers_mx;
  
}
